# Copyright (c) 2003 Sylvain Thenault (thenault@nerim.net)
# Copyright (c) 2003 Logilab
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""Some function usefull to manipulate ast
"""

__author__ = u"Sylvain Thenault"
__revision__ = "$Id: astutils.py,v 1.7 2003/09/22 08:22:05 syt Exp $"

from __future__ import nested_scopes
import symbol
import token
from types import TupleType

# keywords that may appears in a list of tokens
KEYWORDS = ('and', 'or', 'not', 'for', 'in', 'if', 'is', 'lambda')

    
def debuild(ast_tuple):
    """
    reverse ast_tuple to string
    """
    if type(ast_tuple[1]) is TupleType:
        result = ''
        for child in ast_tuple[1:]: 
            result = '%s%s' % (result, debuild(child))
        return result
    else:
        return ast_tuple[1]

def clean(ast_tuple):
    """
    reverse ast tuple to a list of tokens
    merge sequences (token.NAME, token.DOT, token.NAME)
    """
    result = []
    last = None
    for couple in _clean(ast_tuple):
        if couple[0] == token.NAME and last == token.DOT:
            result[-1][1] += couple[1]
        elif couple[0] == token.DOT and last == token.NAME:
            result[-1][1] += couple[1]
        else:
            result.append(couple)
        last = couple[0]
    return result

def _clean(ast_tuple):
    """ transform the ast into as list of tokens (i.e. final elements)
    """
    if type(ast_tuple[1]) is TupleType:
        v = []
        for c in ast_tuple[1:]:
                v += _clean(c)
        return v
    else:
        return [list(ast_tuple[:2])]
    
def cvrtr(tuple):
    """debug method returning an ast string in a readable fashion"""
    if type(tuple) is TupleType:
        try:
            try:
                txt = 'token.'+token.tok_name[tuple[0]]
            except:
                txt = 'symbol.'+symbol.sym_name[tuple[0]]
        except:
            txt =  'Unknown token/symbol'
        return [txt] + map(cvrtr, tuple[1:])
    else:
        return tuple


def get_names(ast_tuple):
    return get_names_from_tokens(clean(ast_tuple))

def get_names_from_tokens(tokens):
    """
    get a list of names from a list of tokens (i.e comma separated values)
    """
    result = []
    new = None
    skip = 0
    for type, tok in tokens:
        if type == token.NAME:
            if skip:
                continue
            if new is None:
                result.append(tok)
            else:
                result[-1] += '.%s' % tok
                new = None
        elif type == symbol.trailer:
            new = 1
        elif type == token.LSQB:
            skip += 1
            new = None
        elif type == token.RSQB:
            skip -= 1
    return result


class State:
    def __init__(self):
        self._stack = []
        
    def push(self, state):
        self._stack.append(state)
        
    def pop(self):
        self._stack.pop()
        
    def current(self):
        return self._stack and self._stack[-1] or None
    
# FIXME: rewrite using a states table
def get_rhs_consumed_names(ast_tuple, keywords=KEYWORDS,
                           state=None, ignore=None):
    """ given an ast as tuple, return the list of consumed variables
    (token.NAME)
    
    ast tuple is considered as the rhs part of an assigment
    """
    if state is None:
        state = State()
        ignore = []
    result = []
    ttype = ast_tuple[0]
    ch_state = None
    # dotted name
    if ttype == symbol.trailer:
        if ast_tuple[1][0] == token.DOT:
            state.push('dotted_name')
        else:
            state.push('consider')
        ch_state = 1
    # function call
    elif ttype == symbol.argument:
        state.push('call_args')
        ch_state = 1
        # FIXME
        if len(ast_tuple) > 2 and ast_tuple[2][0] == token.EQUAL:
            state.argument = 1
        else:
            state.argument = 0
    # consider truc in "truc.bidule" but not in "truc" on lhs mode
    elif ttype == symbol.power:
        current_state = state.current()
        if (current_state == 'ignore' and
            len(ast_tuple) > 2 and ast_tuple[2][0] == symbol.trailer):
            state.push('consider')
            ch_state = 1
    # lambda state
    elif ttype == symbol.lambdef:
        state.push('lambda')
        ch_state = 2
        ignore.append([])
    elif ttype == symbol.varargslist:
        if state.current() == 'lambda':
            state.push('lambda_args')
            ch_state = 1
    elif ttype == symbol.test:
        current_state = state.current()
        if current_state == 'lambda_args':
            state.push('default_value')
            ch_state = 1
        elif current_state == 'call_args':
            if not state.argument:
                state.push('default_value')
                ch_state = 1
            state.argument = 0
            
    # for state
    elif ttype == symbol.listmaker:
        state.push('listcomp')
        ch_state = 2
        ignore.append([])
    elif ttype == symbol.list_for:
        state.push('for')
        ch_state = 1
    elif ttype == symbol.testlist_safe:
        if state.current() == 'for':
            state.push('for_in')
            ch_state = 1
    # name consumed
    elif ttype == token.NAME:
        tok = ast_tuple[1]
        if not tok in keywords:
            current_state = state.current()
            if current_state in (None, 'default_value', 'lambda',
                                 'listcomp', 'for_in', 'consider'):
                result.append(tok)
            elif current_state in ('for', 'lambda_args'):
                ignore[-1].append(tok)
                
    for _tuple in ast_tuple[1:]:
        # recurse ?
        if type(_tuple) is type(()):
            result += get_rhs_consumed_names(_tuple, keywords,
                                             state, ignore)
    if ch_state is not None:
        state.pop()
        if ch_state == 2:
            level_ignore = ignore.pop()
            result = filter(lambda x: not x in level_ignore, result)
    return result


def get_lhs_consumed_names(ast_tuple, keywords=KEYWORDS):
    """ given an ast as tuple, return the list of consumed variables
    (token.NAME)
    
    ast tuple is considered as the lhs part of an assigment
    """
    state = State()
    state.push('ignore')
    return get_rhs_consumed_names(ast_tuple, keywords, state=state, ignore=[])

